<?php
// Copyright (C) <2015-present>  <it-novum GmbH>
//
// This file is licensed under the terms of the openITCOCKPIT Enterprise Edition license agreement.
// The license agreement and license key were sent with the order confirmation.

declare(strict_types=1);

namespace CustomalertModule\Command;

use App\Model\Table\ContainersTable;
use Cake\Command\Command;
use Cake\Console\Arguments;
use Cake\Console\ConsoleIo;
use Cake\Console\ConsoleOptionParser;
use Cake\I18n\FrozenTime;
use Cake\ORM\Query;
use Cake\ORM\TableRegistry;
use Cake\Utility\Hash;
use CustomalertModule\Model\Entity\Customalert;
use CustomalertModule\Model\Table\CustomalertsTable;
use itnovum\openITCOCKPIT\Core\FileDebugger;

/**
 * CheckAlerts command.
 *
 * Created using
 * oitc bake command -p CustomalertModule CheckAlerts
 *
 * Execute via
 * oitc CustomalertModule.CheckAlerts
 *
 */
class CheckAlertsCommand extends Command {

    // Define Naemon exit codes
    // https://docs.openitcockpit.io/en/development/new-check-plugin/
    const EXIT_OK = 0;
    const EXIT_WARNING = 1;
    const EXIT_CRITICAL = 2;
    const EXIT_UNKNOWN = 3;

    /**
     * @var string[]
     */
    private $availableStateTypes;

    /**
     * @var string[]
     */
    private $stateTypesAsString;

    public function __construct() {
        parent::__construct();

        $this->availableStateTypes = [
            (string)Customalert::STATE_NEW,
            (string)Customalert::STATE_PROGRESS,
            (string)Customalert::STATE_DONE,
            (string)Customalert::STATE_MANUALLY_CLOSED
        ];

        $this->stateTypesAsString = [
            Customalert::STATE_NEW             => 'New',
            Customalert::STATE_PROGRESS        => 'In progress',
            Customalert::STATE_DONE            => 'Done',
            Customalert::STATE_MANUALLY_CLOSED => 'Manually closed'
        ];
    }

    /**
     * Hook method for defining this command's option parser.
     *
     * @see https://book.cakephp.org/4/en/console-commands/commands.html#defining-arguments-and-options
     * @param \Cake\Console\ConsoleOptionParser $parser The parser to be defined
     * @return \Cake\Console\ConsoleOptionParser The built parser.
     */
    public function buildOptionParser(ConsoleOptionParser $parser): ConsoleOptionParser {
        $parser = parent::buildOptionParser($parser);

        $parser->addOption('container-id', [
            'help'     => __d('oitc_console', 'ID of the container open alerts should be monitored. If empty, all alerts will be considered. There can be multiple container-id arguments'),
            'default'  => null,
            'boolean'  => false,
            'multiple' => true,
            'choices'  => [],
            'required' => false,
        ]);

        $parser->addOption('recursive', [
            'help'     => __d('oitc_console', 'Include alerts of all sub-container.'),
            'short'    => 'r',
            'default'  => false,
            'boolean'  => true,
            'multiple' => false,
            'choices'  => [],
            'required' => false,
        ]);

        $parser->addOption('statetype', [
            'help'     => __d('oitc_console', '0=New, 1=In Progress, 2=Done, 3=Manually Closed. There can be multiple statetype arguments'),
            'default'  => Customalert::STATE_NEW,
            'boolean'  => false,
            'multiple' => true,
            'choices'  => $this->availableStateTypes,
            'required' => false,
        ]);

        $parser->addOption('warning', [
            'help'     => __d('oitc_console', 'Number of alerts before the plugin will output a warning state. There can be multiple warning arguments for each --statetype. It is important to have the same order e.g.: --statetype 0 --warning 5 --statetype 1 --warning 10'),
            'short'    => 'w',
            'default'  => '5',
            'boolean'  => false,
            'multiple' => true,
            'required' => false,
        ]);

        $parser->addOption('critical', [
            'help'     => __d('oitc_console', 'Number of alerts before the plugin will output a critical state. There can be multiple critical arguments for each --statetype. It is important to have the same order e.g.: --statetype 0 --critical 10 --statetype 1 --critical 15'),
            'short'    => 'c',
            'default'  => '10',
            'boolean'  => false,
            'multiple' => true,
            'required' => false,
        ]);

        $parser->addOption('age-warning', [
            'help'     => __d('oitc_console', 'Alerts older than the provided number in days will trigger a warning state. There can be multiple age-warning arguments for each --statetype. It is important to have the same order e.g.: --statetype 0 --age-warning 2 --statetype 1 --age-warning 5'),
            'default'  => null,
            'boolean'  => false,
            'multiple' => true,
            'required' => false,
        ]);

        $parser->addOption('age-critical', [
            'help'     => __d('oitc_console', 'Alerts older than the provided number in days will trigger a critical state. There can be multiple age-critical arguments for each --statetype. It is important to have the same order e.g.: --statetype 0 --age-critical 4 --statetype 1 --age-critical 10'),
            'default'  => null,
            'boolean'  => false,
            'multiple' => true,
            'required' => false,
        ]);

        $parser->addOption('message', [
            'help'     => __d('oitc_console', 'Only consider alerts where the message contains provided string'),
            'short'    => 'm',
            'default'  => null,
            'boolean'  => false,
            'multiple' => false,
            'required' => false,
        ]);

        $parser->addOption('long-output', [
            'help'     => __d('oitc_console', 'Determines if the plugin should print detailed information as long output.'),
            'short'    => 'l',
            'default'  => false,
            'boolean'  => true,
            'multiple' => false,
            'choices'  => [],
            'required' => false,
        ]);

        return $parser;
    }

    /**
     * Implement this method with your command's logic.
     *
     * @param \Cake\Console\Arguments $args The command arguments.
     * @param \Cake\Console\ConsoleIo $io The console io
     * @return null|void|int The exit code or null for success
     */
    public function execute(Arguments $args, ConsoleIo $io) {
        try {
            $statetypesWithThresholds = $this->getStatetypesWithThresholds($args);
        } catch (\RuntimeException $e) {
            $io->error($e->getMessage());
            exit(self::EXIT_UNKNOWN);
        }

        // Filter options
        $containerIds = $this->makeArray(
            $args->getOption('container-id')
        );
        $recursive = $args->getOption('recursive');
        $message = $args->getOption('message');

        $printLongoutput = $args->getOption('long-output');

        $resolvedContainerIds = [];
        if (!empty($containerIds) && $recursive) {
            /** @var ContainersTable $ContainersTable */
            $ContainersTable = TableRegistry::getTableLocator()->get('Containers');

            // Lookup all child containers
            foreach ($containerIds as $containerIdToResolve) {
                //get recursive container ids
                $containerIdToResolve = (int)$containerIdToResolve;

                $children = $ContainersTable->getChildren($containerIdToResolve);
                $childrenContainerIds = Hash::extract($children, '{n}.id');
                $recursiveContainerIds = [$containerIdToResolve];
                foreach ($childrenContainerIds as $containerId) {
                    $recursiveContainerIds[] = $containerId;
                }
                $resolvedContainerIds = array_merge($resolvedContainerIds, $recursiveContainerIds);
            }
            $resolvedContainerIds = array_unique($resolvedContainerIds);
        }

        // Verbose information
        // This should help users of this plugin to understand what filters are used
        if ($args->getOption('verbose')) {
            $io->verbose('Passed Arguments');
            $io->verbose('================', 2);
            $io->verbose('Alert states with thresholds:');
            foreach ($statetypesWithThresholds as $statetype) {
                $io->verbose(sprintf('Statetype: %s (%d)', $statetype['statetype_string'], $statetype['statetype']));
                $io->verbose(sprintf('   Warning: >= %s', $statetype['warning']));
                $io->verbose(sprintf('   Critical: >= %s', $statetype['critical']));
                $io->verbose(sprintf('   Age warning: %s (days)', $statetype['age-warning']));
                $io->verbose(sprintf('   Age critical: %s (days)', $statetype['age-critical']));
            }
            $io->verbose(sprintf('Passed container IDs: %s', json_encode($containerIds)));
            $io->verbose('Recursive container lookup is ' . ($recursive ? 'enabled' : 'disabled'));
            $io->verbose('Resolved container IDs ' . json_encode($resolvedContainerIds));
            $io->verbose(sprintf('Custom alert message must contain: "%s"', $message));
            $io->verbose('================', 2);
        }

        /** @var CustomalertsTable $CustomalertsTable */
        $CustomalertsTable = TableRegistry::getTableLocator()->get('CustomalertModule.Customalerts');

        if (!empty($resolvedContainerIds)) {
            // Plugin is running with --recursive parameter
            // Overwrite --container-id containers with resolved container Ids
            $containerIds = $resolvedContainerIds;
        }

        foreach ($statetypesWithThresholds as $index => $statetype) {
            // Select custom alerts per statetype (New, In Progress, Done, Manually closed)
            $query = $CustomalertsTable->find()
                ->select([
                    'Customalerts.id',
                    'Customalerts.service_id',
                    'Customalerts.message',
                    'Customalerts.state',
                    'Customalerts.created',
                    'Customalerts.modified'
                ])
                ->where([
                    'Customalerts.state' => $statetype['statetype']
                ]);

            $query->contain('Services', function (Query $query) use ($containerIds) {
                $query
                    ->disableAutoFields()
                    ->select([
                        'Services.id',
                        'Services.host_id'
                    ])
                    ->contain('Hosts', function (Query $query) use ($containerIds) {
                        $query
                            ->select([
                                'Hosts.id',
                                'Hosts.container_id'
                            ])
                            ->innerJoinWith('HostsToContainersSharing', function (Query $q) use ($containerIds) {
                                if (!empty($containerIds)) {
                                    return $q->where(['HostsToContainersSharing.id IN' => $containerIds]);
                                }
                                return $q;
                            })->innerJoinWith('Containers');
                        return $query;
                    });
                return $query;
            });

            if (!empty($message)) {
                $query->andWhere([
                    'Customalerts.message LIKE' => '%' . $message . '%'
                ]);
            }

            $query->group([
                'Customalerts.id'
            ]);

            $query->disableHydration();
            $customalerts = $query->all();

            $totalAlerts = count($customalerts);
            $ageWarningCount = 0;
            $ageCriticalCount = 0;

            if ($statetype['age-warning'] && $statetype['age-critical']) {
                $warningTime = new \DateTime('now');
                $warningTime->modify(sprintf('-%s day', $statetype['age-warning']));
                $warningTimestamp = $warningTime->getTimestamp();

                $criticalTime = new \DateTime('now');
                $criticalTime->modify(sprintf('-%s day', $statetype['age-critical']));
                $criticalTimestamp = $criticalTime->getTimestamp();

                foreach ($customalerts as $customalert) {
                    /** @var FrozenTime $customalert ['created'] */
                    /** @var FrozenTime $customalert ['modified'] */

                    /*
                    debug([
                        'Warning',
                        'alert_created'     => date('d.m.Y H:i:s', $customalert['created']->getTimestamp()),
                        'warning_timestamp' => date('d.m.Y H:i:s', $warningTimestamp)
                    ]);
                    */
                    if ($customalert['created']->getTimestamp() <= $criticalTimestamp || $customalert['modified']->getTimestamp() <= $criticalTimestamp) {
                        $ageCriticalCount++;
                        continue;
                    }

                    if ($customalert['created']->getTimestamp() <= $warningTimestamp || $customalert['modified']->getTimestamp() <= $warningTimestamp) {
                        $ageWarningCount++;
                    }
                }
            }

            // Store counter values for later evaluations
            $statetypesWithThresholds[$index]['evaluation'] = [
                'totalAlerts'      => $totalAlerts,
                'ageWarningCount'  => $ageWarningCount,
                'ageCriticalCount' => $ageCriticalCount,
            ];
        }

        // Print Naemon output
        // Order by customalert statetype to have always the same order in output
        $statetypesWithThresholds = Hash::sort($statetypesWithThresholds, '{n}.statetype', 'asc');

        $exitCode = self::EXIT_OK;
        $output = '';
        $perfdata = '';
        foreach ($statetypesWithThresholds as $statetype) {
            $output .= sprintf(
                'Customalerts in state %s: %s. ',
                $statetype['statetype_string'],
                $statetype['evaluation']['totalAlerts']
            );

            $gaugeName = str_replace(' ', '_', strtolower($statetype['statetype_string']));
            $perfdata .= sprintf(
                '%s=%d;%d;%d ',
                $gaugeName,
                $statetype['evaluation']['totalAlerts'],
                $statetype['warning'],
                $statetype['critical']
            );

            if ($statetype['age-warning'] && $statetype['age-critical']) {
                $subOutput = '';
                if ($statetype['evaluation']['ageWarningCount'] > 0) {
                    $subOutput .= sprintf(
                        'Warning: %d alerts are older than %d days! ',
                        $statetype['evaluation']['ageWarningCount'],
                        $statetype['age-warning']
                    );

                    if ($exitCode < self::EXIT_WARNING) {
                        $exitCode = self::EXIT_WARNING;
                    }
                }

                if ($statetype['evaluation']['ageCriticalCount'] > 0) {
                    $subOutput .= sprintf(
                        'Critical: %d alerts are older than %d days! ',
                        $statetype['evaluation']['ageCriticalCount'],
                        $statetype['age-critical']
                    );

                    if ($exitCode < self::EXIT_CRITICAL) {
                        $exitCode = self::EXIT_CRITICAL;
                    }
                }

                $output .= $subOutput;
            }

            if ($statetype['evaluation']['totalAlerts'] >= $statetype['warning'] && $exitCode < self::EXIT_WARNING) {
                $exitCode = self::EXIT_WARNING;
            }

            if ($statetype['evaluation']['totalAlerts'] >= $statetype['critical'] && $exitCode < self::EXIT_CRITICAL) {
                $exitCode = self::EXIT_CRITICAL;
            }
        }


        // Print short output + performance data
        $overallState = 'Unknown';
        switch ($exitCode) {
            case self::EXIT_OK:
                $overallState = 'Ok';
                break;

            case self::EXIT_WARNING:
                $overallState = 'Warning';
                break;

            case self::EXIT_CRITICAL:
                $overallState = 'Critical';
                break;
        }

        $io->out(sprintf(
            '%s: %s| %s',
            $overallState,
            $output,
            $perfdata
        ));

        if ($printLongoutput) {
            foreach ($statetypesWithThresholds as $statetype) {
                $io->out('Custom alerts state: ' . $statetype['statetype_string']);
                $io->out('==============================');
                $io->out('    Total alerts: ' . $statetype['evaluation']['totalAlerts']);
                if ($statetype['age-warning'] && $statetype['age-critical']) {
                    $io->out(sprintf('    Warning alerts older than %s days: %s', $statetype['age-warning'], $statetype['evaluation']['ageWarningCount']));
                    $io->out(sprintf('    Critical alerts older than %s days: %s', $statetype['age-critical'], $statetype['evaluation']['ageCriticalCount']));
                }
                $io->out('');
            }
        }

        // Set exit code of the plugin
        return $exitCode;
    }


    /**
     * The option parser has to be a bit more complex than usually because multiple options can occur multiple times
     * like --statetype 0 --warning 2 --critical 3 --statetype 1 --warning 2 --critical 3 means
     * Check for alerts in state 0 (New): Warning when alerts >= 2. Critical when alerts >= 3
     * Check for alerts in state 1 (In Progress): Warning when alerts >= 2. Critical when alerts >= 3
     *
     * @param Arguments $args
     * @return array
     */
    private function getStatetypesWithThresholds(Arguments $args) {
        // Arguments::getOption() returns a string, or string[] depending on if arguments get passed by the user
        // or the default values gets used. For this reason we need to make sure that we always have an array
        $optionStateTypes = $this->makeArray(
            $args->getOption('statetype')
        );

        $optionWarnings = $this->makeArray(
            $args->getOption('warning')
        );

        $optionCriticals = $this->makeArray(
            $args->getOption('critical')
        );

        // Optional options
        $aw = $this->makeArray($args->getOption('age-warning'));
        $ac = $this->makeArray($args->getOption('age-critical'));

        // Make sure all arguments are passed in the same amount
        // oitc CustomalertModule.CheckAlerts --statetype 1 --warning 5 --critical 10 --statetype 0 --warning 10 --critical 20
        if (count($optionStateTypes) === count($optionWarnings) && count($optionWarnings) === count($optionCriticals)) {
            // All arrays have the same length
            // Check optional --age-warning and --age-critical options
            // oitc CustomalertModule.CheckAlerts --statetype 1 --warning 5 -c 10 --statetype 0 --warning 10 -c 1 --age-warning 1 --age-warning 2 --age-critical 1 --age-critical 4
            if (!empty($aw) || !empty($ac)) {
                if (!(count($aw) === count($ac) && count($ac) === count($optionStateTypes))) {
                    // Arrays have not the same count, one argument is missing
                    throw new \RuntimeException(
                        sprintf(
                            "Mismatching amount of arguments passed!\nNumber of --statetype: %d\nNumber of --age-warning: %d\nNumber of --age-critical: %d\n",
                            count($optionStateTypes),
                            count($aw),
                            count($ac)
                        )
                    );
                }
            }

            $statetypesWithThresholds = [];
            foreach ($optionStateTypes as $index => $stateType) {
                $statetypesWithThresholds[] = [
                    'statetype'        => (int)$stateType,
                    'statetype_string' => $this->stateTypesAsString[intval($stateType)],
                    'warning'          => (int)$optionWarnings[$index],
                    'critical'         => (int)$optionCriticals[$index],
                    'age-warning'      => (isset($aw[$index])) ? intval($aw[$index]) : null,
                    'age-critical'     => (isset($ac[$index])) ? intval($ac[$index]) : null,
                ];
            }

            return $statetypesWithThresholds;
        }

        // The arrays have different lengths
        throw new \RuntimeException(
            sprintf(
                "Mismatching amount of arguments passed!\nNumber of --statetype: %d\nNumber of --warning: %d\nNumber of --critical: %d\n",
                count($optionStateTypes),
                count($optionWarnings),
                count($optionCriticals)
            )
        );
    }

    /**
     * Makes sure that the passed value is an array. If not, it will be converted into an array
     * @param mixed $val
     * @return array
     */
    private function makeArray($val) {
        if ($val === null) {
            return [];
        }

        if (!is_array($val)) {
            return [$val];
        }

        return $val;
    }
}
