<?php
// Copyright (C) <2015-present>  <it-novum GmbH>
//
// This file is licensed under the terms of the openITCOCKPIT Enterprise Edition license agreement.
// The license agreement and license key were sent with the order confirmation.

declare(strict_types=1);

namespace CustomalertModule\Command;

use App\Model\Table\ContainersTable;
use App\Model\Table\ServicesTable;
use App\Model\Table\SystemsettingsTable;
use Cake\Command\Command;
use Cake\Console\Arguments;
use Cake\Console\ConsoleIo;
use Cake\Console\ConsoleOptionParser;
use Cake\Core\Configure;
use Cake\Database\Exception\MissingConnectionException;
use Cake\Log\Log;
use Cake\ORM\TableRegistry;
use CustomalertModule\Model\Entity\Customalert;
use CustomalertModule\Model\Entity\CustomalertRule;
use CustomalertModule\Model\Table\CustomalertCommentsTable;
use CustomalertModule\Model\Table\CustomalertRulesTable;
use CustomalertModule\Model\Table\CustomalertsTable;
use itnovum\openITCOCKPIT\Core\DbBackend;
use itnovum\openITCOCKPIT\Core\Servicestatus;
use itnovum\openITCOCKPIT\Core\ServicestatusFields;
use itnovum\openITCOCKPIT\Core\Views\AcknowledgementService;

/**
 * Class CustomAlertWorkerCommand
 *
 * Execute for development or debugging
 * OITC_DEBUG=1 oitc CustomalertModule.custom_alert_worker -v
 *
 * Execute for production
 * systemctl start customalert_worker.service
 */
class CustomAlertWorkerCommand extends Command {

    /**
     * Determines if the current process is the parent process or a forked child process.
     *
     * @var bool
     */
    private $parentProcess = true;

    /**
     * All PIDs of the forked child processes
     *
     * @var array
     */
    private $childPids = [];

    /**
     * @var string
     */
    private $pidfile = '/var/run/oitc_custom_alert_worker.pid';

    /**
     * Amount of running custom_alert_worker processes
     * This is hardcoded to 1 at the moment.
     * @var int
     */
    private $maxWorker = 1;

    /**
     * @var array<string,int>
     * A list of service UUIDs where we are going to parse the output into custom alerts
     * All UUIDs inside this array will get processed by this daemon.
     * Everything else gets dropped.
     */
    private $subscribedServiceUuids = [];

    /**
     * @var ConsoleIo
     */
    private $io;

    /**
     * Debugging option --show-dropped-services
     * @var bool
     */
    private $showDroppedServices = false;

    /**
     * Defines the max length of an alert message
     * This limit depends on the database column length
     * @var int
     */
    private $maxLength = 1024;

    /**
     * @var DbBackend
     */
    private $DbBackend;

    /**
     * Hook method for defining this command's option parser.
     *
     * @see https://book.cakephp.org/3.0/en/console-and-shells/commands.html#defining-arguments-and-options
     *
     * @param \Cake\Console\ConsoleOptionParser $parser The parser to be defined
     * @return \Cake\Console\ConsoleOptionParser The built parser.
     */
    public function buildOptionParser(ConsoleOptionParser $parser): ConsoleOptionParser {
        $parser = parent::buildOptionParser($parser);
        $parser->addOptions([
            'show-dropped-services' => ['help' => __d('oitc_console', 'Print UUIDs of dropped services'), 'boolean' => true, 'default' => false],
        ]);

        return $parser;
    }

    /**
     * Implement this method with your command's logic.
     *
     * @param \Cake\Console\Arguments $args The command arguments.
     * @param \Cake\Console\ConsoleIo $io The console io
     * @return null|void|int The exit code or null for success
     */
    public function execute(Arguments $args, ConsoleIo $io) {
        Configure::load('gearman');
        Configure::load('nagios');

        $this->io = $io;
        $this->showDroppedServices = (bool)$args->getOption('show-dropped-services');

        $this->daemonizing();
    }

    public function daemonizing() {
        if (file_exists($this->pidfile)) {
            $pid = trim(file_get_contents($this->pidfile));
            if (is_numeric($pid)) {
                //We got a pid. Is this a custom_alert_worker or is this a pid from an dead process?
                exec('ps -eaf | grep ' . escapeshellarg($pid) . '| grep custom_alert_worker |grep -v grep', $output);

                if (!empty($output)) {
                    Log::error(sprintf('CustomAlertWorker: Pidfile "%s" already exists.', $this->pidfile));
                    exit(3);
                }
            }

            //This is not a gearman_worker process
            Log::info(sprintf('CustomAlertWorker: Deleted orphaned Pidfile "%s". Did the custom_alert_worker died?', $this->pidfile));
            unlink($this->pidfile);

        }

        $pidfd = fopen($this->pidfile, 'w+');
        fwrite($pidfd, (string)getmypid());
        fclose($pidfd);

        declare(ticks=100);
        pcntl_signal(SIGTERM, [$this, 'sig_handler']);
        pcntl_signal(SIGINT, [$this, 'sig_handler']);

        //Disconnect from DB before forking
        /** @var SystemsettingsTable $SystemsettingsTable */
        //$SystemsettingsTable = TableRegistry::getTableLocator()->get('Systemsettings');
        //$connection = $SystemsettingsTable->getConnection();
        //$connection->disconnect();

        //$driver = $connection->getDriver();
        //$driver->disconnect();

        //TableRegistry::getTableLocator()->destroy('Systemsettings');

        for ($i = 1; $i < $this->maxWorker; $i++) {
            $this->io->info(__('CustomAlertWorker: Fork new worker child'));
            $pid = pcntl_fork();
            if (!$pid) {
                //I am a child process
                $this->parentProcess = false;
                $SystemsettingsTable = TableRegistry::getTableLocator()->get('Systemsettings');

                $this->loop();
            }
            $this->childPids[] = $pid;
            unset($pid);
        }

        $SystemsettingsTable = TableRegistry::getTableLocator()->get('Systemsettings');
        pcntl_signal(SIGCHLD, [$this, 'sig_handler']);
        $this->loop();
    }

    /**
     * Loops until receives SIGTERM or SIGINT
     */
    public function loop() {
        $gearmanConfig = Configure::read('gearman');

        $worker = new \GearmanWorker();
        $worker->addServer($gearmanConfig['address'], $gearmanConfig['port']);
        $timeout = $gearmanConfig['timeout'] ?? 1000;
        $worker->setTimeout($timeout);

        $worker->addFunction('statusngin_ocsp', [$this, 'runJob']);
        $worker->addOptions(GEARMAN_WORKER_NON_BLOCKING);

        $this->subscribedServiceUuids = $this->getServiceUuidsForSubscription();
        $this->io->info(sprintf('I\'m subscribed to %s services', sizeof($this->subscribedServiceUuids)));

        $this->DbBackend = new DbBackend();

        while (true) {
            pcntl_signal_dispatch();

            $worker->work();

            //Block for $gearmanConfig['timeout'] if there are no jobs
            @$worker->wait();
        }
    }

    /**
     * @param \GearmanJob $job
     * @return string
     * @throws \Exception
     */
    public function runJob($job) {
        $payload = $job->workload();
        $payload = json_decode($payload, true);

        // Convert message into a bulk message.
        // It is recommended to enable bulk messages in the statusengine.toml for this queue
        if (!isset($payload['messages'])) {
            $payload = [
                'messages' => [
                    $payload
                ],
                'format'   => 'none'
            ];
        }

        if (!is_array($payload) || !isset($payload['messages'])) {
            Log::error('CustomAlertWorkerCommand received corrupted data: ' . (string)$payload);
            return;
        }


        $servicechecksToProcess = [];
        foreach ($payload['messages'] as $servicecheck) {
            $serviceUuid = $servicecheck['servicecheck']['service_description'];
            $processServicecheck = isset($this->subscribedServiceUuids[$serviceUuid]);

            if ($processServicecheck === true) {
                $servicechecksToProcess[] = $servicecheck;
            }


            if ($processServicecheck === true || $this->showDroppedServices === true) {
                $this->io->verbose(sprintf(
                    'Received servicecheck for %s/%s. Process record: %s',
                    $servicecheck['servicecheck']['host_name'],
                    $servicecheck['servicecheck']['service_description'],
                    ($processServicecheck) ? 'Yes' : 'No'
                ));
            }
        }

        $this->handleCustomalerts($servicechecksToProcess);

        unset($payload);
        unset($servicechecksToProcess);
    }

    /**
     * @param array $servicechecks
     * @return void
     * @throws \App\Lib\Exceptions\MissingDbBackendException
     */
    public function handleCustomalerts(array $servicechecks) {
        if (empty($servicechecks)) {
            return;
        }

        $this->testDbConnection();

        /** @var CustomalertsTable $CustomalertsTable */
        $CustomalertsTable = TableRegistry::getTableLocator()->get('CustomalertModule.Customalerts');
        /** @var CustomalertCommentsTable $CustomalertCommentsTable */
        $CustomalertCommentsTable = TableRegistry::getTableLocator()->get('CustomalertModule.CustomalertComments');
        $ServicestatusTable = $this->DbBackend->getServicestatusTable();
        $AcknowledgementServicesTable = $this->DbBackend->getAcknowledgementServicesTable();

        foreach ($servicechecks as $servicecheck) {
            $serviceUuid = $servicecheck['servicecheck']['service_description'];
            $serviceId = $this->subscribedServiceUuids[$serviceUuid];

            // Is status OK(0)? If yes, close all custom alerts of this service
            if ($servicecheck['servicecheck']['state'] === 0) {
                $customalerts = $CustomalertsTable->getCustomalertsForWorker($serviceId, [Customalert::STATE_NEW, Customalert::STATE_PROGRESS]);
                foreach ($customalerts as $customalert) {
                    $CustomalertsTable->closeCustomalert($customalert);
                }
                //Continue with the next service check record from OCSP queue
                continue;
            }

            if ($servicecheck['servicecheck']['state_type'] === 0) {
                // Ignore SOFT state for now
                continue;
            }

            // Status is != OK - Process the output of the check and create/update/delete custom alerts
            $ServicestatusFields = new ServicestatusFields($this->DbBackend);
            $ServicestatusFields
                ->problemHasBeenAcknowledged()
                ->scheduledDowntimeDepth();
            $servicestatus = $ServicestatusTable->byUuid($serviceUuid, $ServicestatusFields);
            $Servicestatus = new Servicestatus($servicestatus['Servicestatus']);
            if ($Servicestatus->isInDowntime()) {
                // Is the service in downtime? If yes, do not create new custom alerts
                $this->io->verbose(sprintf('Service %s is in a scheduled downtime. No alerts will be created', $serviceUuid));
                //Continue with the next service check record from OCSP queue
                continue;
            }


            // Is the service acknowledged? If yes, add the acknowledgement comment to new created custom alerts.
            $acknowledgementComment = null;
            if ($Servicestatus->isAcknowledged()) {
                $acknowledgement = $AcknowledgementServicesTable->byServiceUuid($serviceUuid);
                if (!empty($acknowledgement)) {
                    $Acknowledgement = new AcknowledgementService($acknowledgement);
                    $acknowledgementComment = $Acknowledgement->getCommentData();
                }
            }

            $outputStr = $servicecheck['servicecheck']['output'];
            if (!empty($servicecheck['servicecheck']['long_output'])) {
                $outputStr = $servicecheck['servicecheck']['long_output'];
            }

            // Naemon replaces a new-line \n with an escaped new line \\n in escape_plugin_output() utils.c for long_output
            // https://github.com/naemon/naemon-core/blob/master/src/naemon/utils.c#L1213-L1218
            $lines = explode("\\n", $outputStr);

            foreach ($lines as $i => $line) {
                $lines[$i] = $this->truncateAlert($line);

                if (empty($line)) {
                    unset($lines[$i]);
                }
            }

            $lines = array_unique($lines);

            $customalerts = $CustomalertsTable->getCustomalertsForWorker($serviceId, [Customalert::STATE_NEW, Customalert::STATE_PROGRESS]);
            $compareArray = [];
            foreach ($customalerts as $customalert) {
                $compareArray[] = $customalert->get('message');
            }

            $customalertsToClose = array_diff($compareArray, $lines);
            $customalertsToCreate = array_diff($lines, $compareArray);

            foreach ($customalertsToCreate as $alertString) {
                $customalert = $CustomalertsTable->newEntity([
                    'service_id'   => $serviceId,
                    'service_uuid' => $serviceUuid,
                    'message'      => $alertString,
                    'state'        => Customalert::STATE_NEW
                ]);
                $CustomalertsTable->createCustomalert($customalert);

                if (!empty($acknowledgementComment)) {
                    // Service has an acknowledgement.
                    // Add the acknowledgement comment as annotation for all new alerts.
                    $comment = $CustomalertCommentsTable->newEntity([
                        'customalert_id' => $customalert->id,
                        'user_id'        => 0,
                        'comment'        => $acknowledgementComment,
                        'entry_time'     => date('Y-m-d H:i:s'),
                    ]);
                    $CustomalertCommentsTable->save($comment);
                    $CustomalertsTable->changeState(
                        $customalert,
                        Customalert::STATE_PROGRESS,
                        $comment->id
                    );
                }
            }

            foreach ($customalertsToClose as $alertString) {
                foreach ($customalerts as $customalert) {
                    if ($customalert->get('message') === $alertString) {
                        $CustomalertsTable->closeCustomalert($customalert);
                    }
                }
            }

            $this->io->verbose(sprintf(
                'Created %s alerts. Closed %s alerts.',
                sizeof($customalertsToCreate),
                sizeof($customalertsToClose)
            ));
        }
    }

    /**
     * Avoid "MySQL server has gone away"
     * @param bool $isRecursion
     * @return bool
     */
    public function testDbConnection(bool $isRecursion = false): bool {
        try {
            $SystemsettingsTable = TableRegistry::getTableLocator()->get('Systemsettings');
            $result = $SystemsettingsTable->find()->first();
        } catch (\PDOException $e) {
            $sqlErrorState = $e->errorInfo[0]; //HY000
            $errorCode = $e->errorInfo[1]; //2006
            $errorString = $e->errorInfo[2]; //MySQL server has gone away
            //$e->getMessage() return 'SQLSTATE[HY000]: General error: 2006 MySQL server has gone away'

            $this->io->error('CustomAlertWorkerCommand database error ' . $e->getMessage());

            if ($isRecursion === true) {
                $this->io->error('Unable to reconnect to database');
                return false;
            }

            if ($errorString === 'MySQL server has gone away') {
                $this->io->info('Trying to reconnect to database to recover...');
                TableRegistry::getTableLocator()->destroy('Systemsettings');
                $this->testDbConnection(true);
            }
        } catch (MissingConnectionException $e) {
            $this->io->error('CustomAlertWorkerCommand database error ' . $e->getMessage());

            if ($isRecursion === true) {
                $this->io->error('Unable to reconnect to database');
                return false;
            }

            $this->io->info('Trying to reconnect to database to recover...');
            sleep(5);
            TableRegistry::getTableLocator()->destroy('Systemsettings');
            $this->testDbConnection(true);
        }

        return true;
    }

    /**
     * The signal handler is called by the linux kernel or pcntl_signal_dispatch and will handel singals - obviously
     * Kill forked child processes, delete pidfile and exit on SIGTERM and SIGINT
     *
     * @param int $signo , the signal caught by pcntl_signal_dispatch()
     *
     */
    public function sig_handler($signo) {
        switch ($signo) {
            case SIGTERM:
            case SIGINT:
                if ($this->parentProcess) {
                    // I'm the parent process
                    // Killing all my child processes :'(
                    foreach ($this->childPids as $pid) {
                        posix_kill($pid, SIGTERM);
                    }

                    foreach ($this->childPids as $pid) {
                        pcntl_waitpid($pid, $status);
                    }

                    if (file_exists($this->pidfile)) {
                        unlink($this->pidfile);
                    }
                }
                exit(0);
                break;

            case SIGCHLD:
                // Child died
                // Get the dead child pid and clean up the zombie process
                if ($this->parentProcess) {
                    $deadChildPid = pcntl_wait($status, WNOHANG);
                    $pids = [];
                    foreach ($this->childPids as $pid) {
                        if ($pid != $deadChildPid) {
                            $pids[] = $pid;
                        }
                    }
                    $this->childPids = $pids;
                    unset($pids);
                }
                break;
        }
    }

    public function truncateAlert(string $alert) {
        return substr($alert, 0, $this->maxLength);
    }

    public function getServiceUuidsForSubscription(): array {
        /** @var CustomalertRulesTable $CustomalertRulesTable */
        $CustomalertRulesTable = TableRegistry::getTableLocator()->get('CustomalertModule.CustomalertRules');
        /** @var ServicesTable $ServicesTable */
        $ServicesTable = TableRegistry::getTableLocator()->get('Services');
        /** @var ContainersTable $ContainersTable */
        $ContainersTable = TableRegistry::getTableLocator()->get('Containers');


        $customalertRules = $CustomalertRulesTable->find()->all();

        if (empty($customalertRules)) {
            return [];
        }

        $services = [];
        /** @var CustomalertRule $customalertRule */
        foreach ($customalertRules as $customalertRule) {
            $containerIds = [$customalertRule->container_id];

            if ($customalertRule->recursive) {
                if ($customalertRule->container_id === ROOT_CONTAINER) {
                    // This rule apply to all objects!
                    $containerIds = [];
                } else {
                    $tmpContainerIds = $ContainersTable->resolveChildrenOfContainerIds($customalertRule->container_id, false);
                    $containerIds = $ContainersTable->removeRootContainer($tmpContainerIds);
                }
            }

            $servicesFromRule = $CustomalertRulesTable->getServicesByCustomalertRule($ServicesTable, $customalertRule, $containerIds);

            foreach ($servicesFromRule as $serviceUuid => $serviceId) {
                $services[$serviceUuid] = $serviceId;
            }
        }

        return $services;
    }

}
