<?php
// Copyright (C) <2015-present>  <it-novum GmbH>
//
// This file is licensed under the terms of the openITCOCKPIT Enterprise Edition license agreement.
// The license agreement and license key were sent with the order confirmation.

declare(strict_types=1);

namespace CustomalertModule\Controller;

use App\Model\Table\ContainersTable;
use App\Model\Table\ServicesTable;
use Cake\Http\Exception\MethodNotAllowedException;
use Cake\Http\Exception\NotFoundException;
use Cake\ORM\TableRegistry;
use CustomalertModule\Model\Entity\CustomalertRule;
use CustomalertModule\Model\Table\CustomalertRulesTable;
use itnovum\openITCOCKPIT\Database\PaginateOMat;
use itnovum\openITCOCKPIT\Filter\GenericFilter;

/**
 * CustomalertRules Controller
 *
 * @method CustomalertRule[]|\Cake\Datasource\ResultSetInterface paginate($object = null, array $settings = [])
 */
class CustomalertRulesController extends AppController {

    public function index() {
        if (!$this->isApiRequest()) {
            //Only ship template
            return;
        }

        /** @var ContainersTable $ContainersTable */
        $ContainersTable = TableRegistry::getTableLocator()->get('Containers');

        $GenericFilter = new GenericFilter($this->request);
        $GenericFilter->setFilters([
            'like' => [
                'CustomalertRules.name',
                'CustomalertRules.description'
            ]
        ]);

        $PaginateOMat = new PaginateOMat($this, $this->isScrollRequest(), $GenericFilter->getPage());

        $MY_RIGHTS = [];
        if ($this->hasRootPrivileges === false) {
            $MY_RIGHTS = $this->MY_RIGHTS;
        }

        /** @var CustomalertRulesTable $CustomalertRulesTable */
        $CustomalertRulesTable = TableRegistry::getTableLocator()->get('CustomalertModule.CustomalertRules');

        $customalertRules = $CustomalertRulesTable->getCustomalertRulesIndex($GenericFilter, $PaginateOMat, $MY_RIGHTS);
        foreach ($customalertRules as $index => $customalertRule) {
            $customalertRules[$index]['container'] = '/' . $ContainersTable->treePath($customalertRule['container_id']);
            $customalertRules[$index]['allowEdit'] = $this->isWritableContainer($customalertRule['container_id']);

        }

        $this->set('customalertRules', $customalertRules);
        $this->viewBuilder()->setOption('serialize', ['customalertRules']);

    }

    public function add() {
        if (!$this->isApiRequest()) {
            //Only ship template
            return;
        }
        /** @var CustomalertRulesTable $CustomalertRulesTable */
        $CustomalertRulesTable = TableRegistry::getTableLocator()->get('CustomalertModule.CustomalertRules');
        if ($this->request->is('post') && $this->isAngularJsRequest()) {
            $customalertRule = $CustomalertRulesTable->newEmptyEntity();
            $requestData = $this->request->getData();

            $customalertRule = $CustomalertRulesTable->patchEntity($customalertRule, $requestData);
            $CustomalertRulesTable->save($customalertRule);
            if ($customalertRule->hasErrors()) {
                $this->response = $this->response->withStatus(400);
                $this->serializeCake4ErrorMessage($customalertRule);
                return;
            } else {
                //No errors
                $this->serializeCake4Id($customalertRule);
            }
            $this->set('customalertRule', $customalertRule);
            $this->viewBuilder()->setOption('serialize', ['customalertRule']);
        }
    }

    /**
     * @param int|null $id
     */
    public function edit($id = null) {
        if (!$this->isApiRequest()) {
            //Only ship template for AngularJs
            return;
        }

        /** @var CustomalertRulesTable $CustomalertRulesTable */
        $CustomalertRulesTable = TableRegistry::getTableLocator()->get('CustomalertModule.CustomalertRules');
        if (!$CustomalertRulesTable->existsById($id)) {
            throw new NotFoundException(__('Custom alert rule not found'));
        }

        $customalertRule = $CustomalertRulesTable->get($id);
        if (!$this->allowedByContainerId($customalertRule['container_id'])) {
            $this->render403();
            return;
        }

        if ($this->request->is('get') && $this->isAngularJsRequest()) {
            $customalertRule['allowEdit'] = $this->isWritableContainer($customalertRule['container_id']);

            $this->set('customalertRule', $customalertRule);
            $this->viewBuilder()->setOption('serialize', ['customalertRule']);
            return;
        }

        if ($this->request->is('post') || $this->request->is('put')) {
            $requestData = $this->request->getData(null, []);

            $customalertRule = $CustomalertRulesTable->patchEntity($customalertRule, $requestData);

            $CustomalertRulesTable->save($customalertRule);

            if ($customalertRule->hasErrors()) {
                $this->response = $this->response->withStatus(400);
                $this->set('error', $customalertRule->getErrors());
                $this->viewBuilder()->setOption('serialize', ['error']);
                return;
            }

            $this->set('customalertRule', $customalertRule);
            $this->viewBuilder()->setOption('serialize', ['customalertRule']);
        }
    }

    /**
     * @param null $id
     */
    public function delete($id = null) {
        if (!$this->request->is('post')) {
            throw new MethodNotAllowedException();
        }

        /** @var CustomalertRulesTable $CustomalertRulesTable */
        $CustomalertRulesTable = TableRegistry::getTableLocator()->get('CustomalertModule.CustomalertRules');

        if (!$CustomalertRulesTable->existsById($id)) {
            throw new NotFoundException(__('Custom alert rule not found'));
        }

        $customalertRule = $CustomalertRulesTable->get($id);
        if (!$this->allowedByContainerId($customalertRule->get('container_id'))) {
            $this->render403();
            return;
        }

        if ($CustomalertRulesTable->delete($customalertRule)) {
            $this->set('success', true);
            $this->viewBuilder()->setOption('serialize', ['success']);

            return;
        }

        $this->response = $this->response->withStatus(400);
        $this->set('success', false);
        $this->viewBuilder()->setOption('serialize', ['success']);
        return;
    }

    /**
     * @param int|null $id
     */
    public function services($id = null) {
        if (!$this->isApiRequest()) {
            //Only ship template for AngularJs
            return;
        }

        /** @var CustomalertRulesTable $CustomalertRulesTable */
        $CustomalertRulesTable = TableRegistry::getTableLocator()->get('CustomalertModule.CustomalertRules');

        /** @var ServicesTable $ServicesTable */
        $ServicesTable = TableRegistry::getTableLocator()->get('Services');

        if (!$CustomalertRulesTable->existsById($id)) {
            throw new NotFoundException(__('Custom alert rule not found'));
        }

        $customalertRule = $CustomalertRulesTable->get($id);
        if (!$this->allowedByContainerId($customalertRule['container_id'])) {
            $this->render403();
            return;
        }

        $GenericFilter = new GenericFilter($this->request);
        $GenericFilter->setFilters([
            'like' => [
                'Hosts.name',
                'servicename'
            ]
        ]);

        $PaginateOMat = new PaginateOMat($this, $this->isScrollRequest(), $GenericFilter->getPage());

        $containerIds = [$customalertRule->container_id];
        if ($customalertRule->recursive) {
            /** @var ContainersTable $ContainersTable */
            $ContainersTable = TableRegistry::getTableLocator()->get('Containers');
            if ($customalertRule->container_id === ROOT_CONTAINER) {
                // Alert rule apply for all objects the user can see!
                $containerIds = $this->MY_RIGHTS;
            } else {
                $tmpContainerIds = $ContainersTable->resolveChildrenOfContainerIds($customalertRule->container_id, false);
                $containerIds = $ContainersTable->removeRootContainer($tmpContainerIds);
            }
        }

        if ($this->request->is('get') && $this->isAngularJsRequest()) {
            $services = $CustomalertRulesTable->getServicesForOverviewByCustomalertRule(
                $ServicesTable,
                $customalertRule,
                $PaginateOMat,
                $GenericFilter,
                $containerIds
            );
            $customalertRule['allowEdit'] = $this->isWritableContainer($customalertRule['container_id']);

            $this->set('customalertRule', $customalertRule);
            $this->set('services', $services);
            $this->set('success', true);
            $this->viewBuilder()->setOption('serialize', ['customalertRule', 'services', 'success']);
            return;
        }

        $this->response = $this->response->withStatus(400);
        $this->set('success', false);
        $this->viewBuilder()->setOption('serialize', ['success']);
    }
}
