<?php
// Copyright (C) <2015-present>  <it-novum GmbH>
//
// This file is licensed under the terms of the openITCOCKPIT Enterprise Edition license agreement.
// The license agreement and license key were sent with the order confirmation.

declare(strict_types=1);

namespace CustomalertModule\Controller;

use App\itnovum\openITCOCKPIT\Monitoring\Naemon\ExternalCommands;
use App\Model\Table\ContainersTable;
use App\Model\Table\HostsTable;
use App\Model\Table\ServicesTable;
use Cake\Http\Exception\BadRequestException;
use Cake\Http\Exception\MethodNotAllowedException;
use Cake\Http\Exception\NotFoundException;
use Cake\I18n\FrozenTime;
use Cake\ORM\TableRegistry;
use Cake\Utility\Hash;
use CustomalertModule\Lib\CustomalertConditions;
use CustomalertModule\Lib\CustomalertFilter;
use CustomalertModule\Model\Entity\Customalert;
use CustomalertModule\Model\Table\CustomalertCommentsTable;
use CustomalertModule\Model\Table\CustomalertsTable;
use itnovum\openITCOCKPIT\Core\AngularJS\Api;
use itnovum\openITCOCKPIT\Core\AngularJS\Request\AngularRequest;
use itnovum\openITCOCKPIT\Core\HoststatusFields;
use itnovum\openITCOCKPIT\Core\ValueObjects\User;
use itnovum\openITCOCKPIT\Core\Views\ContainerPermissions;
use itnovum\openITCOCKPIT\Database\PaginateOMat;
use itnovum\openITCOCKPIT\Filter\GenericFilter;

/**
 *
 */
class CustomalertsController extends AppController {
    /**
     * Index method
     *
     * @return \Cake\Http\Response|null|void Renders view
     */
    public function index() {
        if (!$this->isApiRequest()) {
            //Only ship template
            return;
        }

        $User = new User($this->getUser());
        $UserTime = $User->getUserTime();

        $AngularRequest = new AngularRequest($this->request);
        $CustomalertFilter = new CustomalertFilter($this->request);
        $PaginateOMat = new PaginateOMat($this, $this->isScrollRequest(), $CustomalertFilter->getPage());
        $recursive = $this->request->getQuery('recursive', 'false') === 'true';
        $CustomalertConditions = new CustomalertConditions();
        $CustomalertConditions->setFrom($AngularRequest->getFrom());
        $CustomalertConditions->setTo($AngularRequest->getTo());
        $CustomalertConditions->setOrder($AngularRequest->getOrderForPaginator('Customalerts.created', 'desc'));
        $CustomalertConditions->setRecursiveContainer($recursive);

        $where = $CustomalertFilter->indexFilter();
        if (isset($where['Hosts.container_id IN']) && $CustomalertConditions->isRecursiveContainer()) {
            /** @var ContainersTable $ContainersTable */
            $ContainersTable = TableRegistry::getTableLocator()->get('Containers');

            // Lookup all child containers
            $resolvedContainerIds = [];
            foreach ($where['Hosts.container_id IN'] as $containerIdToResolve) {
                //get recursive container ids
                $containerIdToResolve = (int)$containerIdToResolve;

                $children = $ContainersTable->getChildren($containerIdToResolve);
                $containerIds = Hash::extract($children, '{n}.id');
                $recursiveContainerIds = [$containerIdToResolve];
                foreach ($containerIds as $containerId) {
                    if (in_array($containerId, $this->MY_RIGHTS)) {
                        $recursiveContainerIds[] = $containerId;
                    }
                }
                $resolvedContainerIds = array_merge($resolvedContainerIds, $recursiveContainerIds);
            }
            $resolvedContainerIds = array_unique($resolvedContainerIds);
            $where['Hosts.container_id IN'] = $resolvedContainerIds;
        }

        $CustomalertConditions->setConditions($where);

        $MY_RIGHTS = [];
        if ($this->hasRootPrivileges === false) {
            $MY_RIGHTS = $this->MY_RIGHTS;
        }

        /** @var CustomalertsTable $CustomalertsTable */
        $CustomalertsTable = TableRegistry::getTableLocator()->get('CustomalertModule.Customalerts');

        /** @var HostsTable $HostsTable */
        $HostsTable = TableRegistry::getTableLocator()->get('Hosts');
        $hostContainers = [];

        $customalerts = $CustomalertsTable->getCustomalertsIndex($CustomalertConditions, $PaginateOMat, $MY_RIGHTS);
        $customalertStateOverview = [
            Customalert::STATE_NEW             => 0,
            Customalert::STATE_PROGRESS        => 0,
            Customalert::STATE_DONE            => 0,
            Customalert::STATE_MANUALLY_CLOSED => 0
        ];

        $customalertGlobalOverview = [
            Customalert::STATE_NEW      => 0,
            Customalert::STATE_PROGRESS => 0
        ];

        $customalertsGlobalOverviewNewAndInProgress = $CustomalertsTable->getCustomalertsGlobalOverviewNewAndInProgress(
            $CustomalertConditions,
            $MY_RIGHTS
        );

        foreach ($customalertsGlobalOverviewNewAndInProgress as $customalertGlobalOverviewNewAndInProgress) {
            $customalertGlobalOverview[$customalertGlobalOverviewNewAndInProgress['state']] = (int)$customalertGlobalOverviewNewAndInProgress['count'];
        }

        $customalertGlobalOverview['total'] = array_sum($customalertGlobalOverview);

        foreach ($customalerts as $index => $customalert) {
            $created = new FrozenTime($customalert['created']);
            $modified = new FrozenTime($customalert['modified']);
            $customalerts[$index]['allow_edit'] = $this->hasRootPrivileges;
            $customalerts[$index]['createdInHumanShort'] = $UserTime->secondsInHumanShort(time() - $created->timestamp);
            $customalerts[$index]['modifiedInHumanShort'] = $UserTime->secondsInHumanShort(time() - $modified->timestamp);
            $customalerts[$index]['created'] = $UserTime->format($created);
            $customalerts[$index]['modified'] = $UserTime->format($modified);
            if (!empty($customalert['CustomalertComments']['entry_time'])) {
                $entryTime = new FrozenTime($customalert['modified']);
                $customalerts[$index]['CustomalertComments']['entry_time'] = $UserTime->format($entryTime);
            }
            $customalerts[$index]['modified'] = $UserTime->format($modified);

            // Select last24hoursAlerts for each alert because the sub-select was killing the MySQL server when selecting DONE alerts
            $customalerts[$index]['last24hoursAlerts'] = $CustomalertsTable->getLast24hCounter((int)$customalert['service_id'], $customalert['message']);
            $customalertStateOverview[$customalert['state']]++;
        }
        if ($this->hasRootPrivileges === false) {
            foreach ($customalerts as $index => $customalert) {
                $hostId = $customalert['_matchingData']['Hosts']['id'];
                if (!isset($hostContainers[$hostId])) {
                    $hostContainers[$hostId] = $HostsTable->getHostContainerIdsByHostId($hostId);
                }
                $ContainerPermissions = new ContainerPermissions($this->MY_RIGHTS_LEVEL, $hostContainers[$hostId]);
                $customalerts[$index]['allow_edit'] = $ContainerPermissions->hasPermission();
            }
        }

        // Remove array keys (left over in group view) to get an array [] in javascript (not in object {} to keep sort)
        $customalerts = array_values($customalerts);

        $this->set('customalerts', $customalerts);
        $this->set('customalertStateOverview', $customalertStateOverview);
        $this->set('customalertGlobalOverview', $customalertGlobalOverview);
        $this->viewBuilder()->setOption('serialize', [
            'customalerts',
            'customalertStateOverview',
            'customalertGlobalOverview'
        ]);
    }

    /**
     * View method
     *
     * @param string|null $id Customalert id.
     * @return \Cake\Http\Response|null|void Renders view
     * @throws \Cake\Datasource\Exception\RecordNotFoundException When record not found.
     */
    public function view($id = null) {
        $customalert = $this->Customalerts->get($id, [
            'contain' => [],
        ]);

        $this->set(compact('customalert'));
    }

    /**
     * @param $id
     * @return void
     */
    public function annotate($id = null) {
        if (!$this->isAngularJsRequest()) {
            //Only ship HTML Template
            return;
        }
        /** @var CustomalertsTable $CustomalertsTable */
        $CustomalertsTable = TableRegistry::getTableLocator()->get('CustomalertModule.Customalerts');
        if (!$CustomalertsTable->existsById($id)) {
            throw new NotFoundException();
        }

        $setAnnotationAsAcknowledgement = $this->request->getQuery('setAnnotationAsAcknowledgement', 'false') === 'true';

        $User = new User($this->getUser());

        /** @var CustomalertCommentsTable $CustomalertCommentsTable */
        $CustomalertCommentsTable = TableRegistry::getTableLocator()->get('CustomalertModule.CustomalertComments');
        if ($this->request->is('post') && $this->isAngularJsRequest()) {

            $customalertcommentEntity = $CustomalertCommentsTable->newEmptyEntity();

            $requestData = $this->request->getData();
            $requestData['customalert_id'] = $id;
            $requestData['user_id'] = $User->getId();
            $requestData['entry_time'] = date('Y-m-d H:i:s');

            $customalertcommentEntity = $CustomalertCommentsTable->patchEntity($customalertcommentEntity, $requestData);
            $CustomalertCommentsTable->save($customalertcommentEntity);
            if ($customalertcommentEntity->hasErrors()) {
                $this->response = $this->response->withStatus(400);
                $this->serializeCake4ErrorMessage($customalertcommentEntity);
                return;
            } else {
                //No errors - set customalert record in progress
                $customalert = $CustomalertsTable->get($id);
                $CustomalertsTable->changeState(
                    $customalert,
                    Customalert::STATE_PROGRESS,
                    $customalertcommentEntity->id
                );

                if ($setAnnotationAsAcknowledgement) {
                    // Check if host is down and set the comment if the annotation as acknowledgement for the host.
                    /** @var HostsTable $HostsTable */
                    $HostsTable = TableRegistry::getTableLocator()->get('Hosts');
                    /** @var ServicesTable $ServicesTable */
                    $ServicesTable = TableRegistry::getTableLocator()->get('Services');

                    $service = $ServicesTable->getServiceByIdForPermissionsCheck($customalert->service_id);
                    if ($this->hasRootPrivileges) {
                        $allowEdit = true;
                    } else {
                        $ContainerPermissions = new ContainerPermissions($this->MY_RIGHTS_LEVEL, $service->getContainerIds());
                        $allowEdit = $ContainerPermissions->hasPermission();
                    }

                    $canSubmitExternalCommands = $this->hasPermission('externalcommands', 'hosts', '');

                    if ($allowEdit && $canSubmitExternalCommands) {
                        // User has permissions to submit external commands for this host / service
                        $ExternalCommands = new ExternalCommands();
                        $options = [
                            'hostUuid'     => $service->host->uuid,
                            'author'       => $User->getFullName(),
                            'comment'      => str_replace("\n", ' ', $customalertcommentEntity->comment),
                            'sticky'       => 2, // 2 = sticky with Nagios Core, 1 = Sticky with Naemon Core, 2 works for both :)
                            'type'         => 'hostOnly',
                            'satellite_id' => $service->host->satellite_id
                        ];
                        $ExternalCommands->setHostAck($options);
                    }
                }

                $this->serializeCake4Id($customalertcommentEntity);
            }
            $this->set('customalertcomment', $customalertcommentEntity);
            $this->viewBuilder()->setOption('serialize', ['customalertcomment']);
        }
    }

    /**
     * @param $id
     * @return void
     */
    public function close($id = null) {
        if (!$this->isAngularJsRequest()) {
            //Only ship HTML Template
            return;
        }
        /** @var CustomalertsTable $CustomalertsTable */
        $CustomalertsTable = TableRegistry::getTableLocator()->get('CustomalertModule.Customalerts');
        if (!$CustomalertsTable->existsById($id)) {
            throw new NotFoundException();
        }

        $User = new User($this->getUser());

        /** @var CustomalertCommentsTable $CustomalertCommentsTable */
        $CustomalertCommentsTable = TableRegistry::getTableLocator()->get('CustomalertModule.CustomalertComments');
        if ($this->request->is('post') && $this->isAngularJsRequest()) {

            $customalertcommentEntity = $CustomalertCommentsTable->newEmptyEntity();

            $requestData = $this->request->getData();
            $requestData['customalert_id'] = $id;
            $requestData['user_id'] = $User->getId();
            $requestData['entry_time'] = date('Y-m-d H:i:s');

            $customalertcommentEntity = $CustomalertCommentsTable->patchEntity($customalertcommentEntity, $requestData);
            $CustomalertCommentsTable->save($customalertcommentEntity);
            if ($customalertcommentEntity->hasErrors()) {
                $this->response = $this->response->withStatus(400);
                $this->serializeCake4ErrorMessage($customalertcommentEntity);
                return;
            } else {
                //No errors - set customalert record in progress
                $customalert = $CustomalertsTable->get($id);
                $CustomalertsTable->changeState(
                    $customalert,
                    Customalert::STATE_MANUALLY_CLOSED,
                    $customalertcommentEntity->id
                );

                $this->serializeCake4Id($customalertcommentEntity);
            }
            $this->set('customalertcomment', $customalertcommentEntity);
            $this->viewBuilder()->setOption('serialize', ['customalertcomment']);
        }
    }

    public function mass_annotate() {
        //Return HTML Template for PaginatorDirective
        return;
    }

    public function mass_close() {
        //Return HTML Template for PaginatorDirective
        return;
    }

    /**
     * @param $id
     * @return void
     * @throws \Exception
     */
    public function history($id = null) {
        if (!$this->isApiRequest()) {
            //Only ship template for AngularJs
            return;
        }

        /** @var CustomalertsTable $CustomalertsTable */
        $CustomalertsTable = TableRegistry::getTableLocator()->get('CustomalertModule.Customalerts');


        if (!$CustomalertsTable->existsById($id)) {
            throw new NotFoundException(__('Custom alert not found'));
        }

        $history = $CustomalertsTable->getCustomalertHistoryById($id);

        if (!$this->allowedByContainerId($history['service']['host']['container_id'])) {
            $this->render403();
            return;
        }
        $User = new User($this->getUser());
        $UserTime = $User->getUserTime();

        if ($this->request->is('get') && $this->isAngularJsRequest()) {
            $customalert = [
                'service'     => $history['service'],
                'servicename' => $history['servicename'],
                'message'     => $history['message']
            ];
            foreach ($history['customalert_statehistory'] as $historyRecord) {
                $stateTime = new FrozenTime($historyRecord['state_time']);
                $stateHistory[] = [
                    'state_time'          => $UserTime->format($stateTime),
                    'timeAgoInWords'      => $UserTime->secondsInHumanShort(time() - $stateTime->timestamp),
                    'user'                => !empty($historyRecord['customalert_comment']['user']['id']),
                    'customalert_comment' => $historyRecord['customalert_comment'],
                    'state'               => $historyRecord['state']
                ];
            }


            $this->set('history', $stateHistory);
            $this->set('customalert', $customalert);
            $this->set('success', true);
            $this->viewBuilder()->setOption('serialize', ['customalert', 'history', 'success']);
            return;
        }

        $this->response = $this->response->withStatus(400);
        $this->set('success', false);
        $this->viewBuilder()->setOption('serialize', ['success']);
    }

    /**
     * @param $id
     * @return void
     * @throws \Exception
     */
    public function service_history($serviceId = null) {
        if (!$this->isApiRequest()) {
            //Only ship template for AngularJs
            return;
        }

        /** @var CustomalertsTable $CustomalertsTable */
        $CustomalertsTable = TableRegistry::getTableLocator()->get('CustomalertModule.Customalerts');

        /** @var ServicesTable $ServicesTable */
        $ServicesTable = TableRegistry::getTableLocator()->get('Services');


        if (!$ServicesTable->existsById($serviceId)) {
            throw new NotFoundException(__('Service not found'));
        }

        $Filter = new GenericFilter($this->request);
        $PaginateOMat = new PaginateOMat($this, $this->isScrollRequest(), $Filter->getPage());

        $customalertHistory = $CustomalertsTable->getCustomalertHistoryByServiceId($serviceId, $PaginateOMat);
        $service = $ServicesTable->get($serviceId, [
            'contain' => [
                'Hosts'
            ]
        ]);

        if (!$this->allowedByContainerId($service->get('host')->get('container_id'))) {
            $this->render403();
            return;
        }
        $User = new User($this->getUser());
        $UserTime = $User->getUserTime();

        $stateHistory = [];

        if ($this->request->is('get') && $this->isAngularJsRequest()) {
            foreach ($customalertHistory as $key => $customalert) {
                $modified = new FrozenTime($customalert['modified']);
                $stateHistory[$key]['message'] = $customalert['message'];
                $stateHistory[$key]['state'] = $customalert['state'];
                $stateHistory[$key]['modified'] = $UserTime->format($modified);
                $stateHistory[$key]['timeAgoInWords'] = $UserTime->secondsInHumanShort(time() - $modified->timestamp);
                foreach ($customalert['customalert_statehistory'] as $stateRecord) {
                    $stateTime = new FrozenTime($stateRecord['state_time']);
                    $stateHistory[$key]['customalert_statehistory'][] =
                        [
                            'state_time'          => $UserTime->format($stateTime),
                            'timeAgoInWords'      => $UserTime->secondsInHumanShort(time() - $stateTime->timestamp),
                            'user'                => !empty($stateRecord['customalert_comment']['user']['id']),
                            'customalert_comment' => $stateRecord['customalert_comment'],
                            'state'               => $stateRecord['state']
                        ];
                }
            }

            $this->set('history', $stateHistory);
            $this->set('success', true);
            $this->viewBuilder()->setOption('serialize', ['history', 'success']);
            return;
        }

        $this->response = $this->response->withStatus(400);
        $this->set('success', false);
        $this->viewBuilder()->setOption('serialize', ['success']);
    }

    /****************************
     *       AJAX METHODS       *
     ****************************/

    public function loadContainers() {
        if (!$this->isApiRequest()) {
            //Only ship template for AngularJs
            return;
        }

        /** @var $ContainersTable ContainersTable */
        $ContainersTable = TableRegistry::getTableLocator()->get('Containers');

        if ($this->hasRootPrivileges === true) {
            $containers = $ContainersTable->easyPath($this->MY_RIGHTS, OBJECT_HOST, [], $this->hasRootPrivileges, [CT_HOSTGROUP]);
        } else {
            $containers = $ContainersTable->easyPath($this->getWriteContainers(), OBJECT_HOST, [], $this->hasRootPrivileges, [CT_HOSTGROUP]);
        }

        $this->set('containers', Api::makeItJavaScriptAble($containers));
        $this->viewBuilder()->setOption('serialize', ['containers']);

    }

    public function checkHoststatusForAcknowledgements() {
        if (!$this->request->is('post')) {
            throw new MethodNotAllowedException();
        }

        /** @var HostsTable $HostsTable */
        $HostsTable = TableRegistry::getTableLocator()->get('Hosts');
        $HoststatusTable = $this->DbBackend->getHoststatusTable();

        $hostIds = $this->request->getData('hostIds', []);
        if (!is_array($hostIds)) {
            throw new BadRequestException('hostIds needs to be an array');
        }

        $allHosts = [];
        foreach ($hostIds as $hostId) {
            if (is_numeric($hostId)) {
                $host = $HostsTable->getHostByIdForPermissionCheck($hostId);
                if (!empty($host)) {
                    if ($this->hasRootPrivileges) {
                        $allowEdit = true;
                    } else {
                        $ContainerPermissions = new ContainerPermissions($this->MY_RIGHTS_LEVEL, $host->getContainerIds());
                        $allowEdit = $ContainerPermissions->hasPermission();
                    }

                    $HoststatusFields = new HoststatusFields($this->DbBackend);
                    $HoststatusFields->currentState();
                    $hoststatus = $HoststatusTable->byUuid($host->uuid, $HoststatusFields);

                    if (!empty($hoststatus)) {
                        if ($hoststatus['Hoststatus']['current_state'] > 0) {
                            $allHosts[] = [
                                'id'            => $host->id,
                                'name'          => $host->name,
                                'current_state' => $hoststatus['Hoststatus']['current_state'],
                                'allowEdit'     => $allowEdit
                            ];
                        }
                    }

                }
            }
        }

        $canSubmitExternalCommands = $this->hasPermission('externalcommands', 'hosts', '');
        $this->set('all_hosts', $allHosts);
        $this->set('canSubmitExternalCommands', $canSubmitExternalCommands);
        $this->viewBuilder()->setOption('serialize', [
            'all_hosts',
            'canSubmitExternalCommands'
        ]);
    }

}
