<?php
declare(strict_types=1);

namespace CustomalertModule\Model\Table;

use App\Lib\Traits\PaginationAndScrollIndexTrait;
use App\Model\Table\ContainersTable;
use App\Model\Table\ServicesTable;
use Cake\Database\Expression\ComparisonExpression;
use Cake\ORM\RulesChecker;
use Cake\ORM\Table;
use Cake\Utility\Hash;
use Cake\Validation\Validator;
use CustomalertModule\Model\Entity\CustomalertRule;
use itnovum\openITCOCKPIT\Database\PaginateOMat;
use itnovum\openITCOCKPIT\Filter\GenericFilter;

/**
 * CustomalertRules Model
 *
 * @property ContainersTable $Containers
 *
 * @method CustomalertRule newEmptyEntity()
 * @method CustomalertRule newEntity(array $data, array $options = [])
 * @method CustomalertRule[] newEntities(array $data, array $options = [])
 * @method CustomalertRule get($primaryKey, $options = [])
 * @method CustomalertRule findOrCreate($search, ?callable $callback = null, $options = [])
 * @method CustomalertRule patchEntity(\Cake\Datasource\EntityInterface $entity, array $data, array $options = [])
 * @method CustomalertRule[] patchEntities(iterable $entities, array $data, array $options = [])
 * @method CustomalertRule|false save(\Cake\Datasource\EntityInterface $entity, $options = [])
 * @method CustomalertRule saveOrFail(\Cake\Datasource\EntityInterface $entity, $options = [])
 * @method CustomalertRule[]|\Cake\Datasource\ResultSetInterface|false saveMany(iterable $entities, $options = [])
 * @method CustomalertRule[]|\Cake\Datasource\ResultSetInterface saveManyOrFail(iterable $entities, $options = [])
 * @method CustomalertRule[]|\Cake\Datasource\ResultSetInterface|false deleteMany(iterable $entities, $options = [])
 * @method CustomalertRule[]|\Cake\Datasource\ResultSetInterface deleteManyOrFail(iterable $entities, $options = [])
 *
 * @mixin \Cake\ORM\Behavior\TimestampBehavior
 */
class CustomalertRulesTable extends Table {
    use PaginationAndScrollIndexTrait;

    /**
     * Initialize method
     *
     * @param array $config The configuration for the Table.
     * @return void
     */
    public function initialize(array $config): void {
        parent::initialize($config);

        $this->setTable('customalert_rules');
        $this->setDisplayField('name');
        $this->setPrimaryKey('id');

        $this->addBehavior('Timestamp');

        $this->belongsTo('Containers', [
            'foreignKey' => 'container_id',
            'joinType'   => 'INNER',
            'className'  => 'Containers',
        ]);
    }

    /**
     * Default validation rules.
     *
     * @param \Cake\Validation\Validator $validator Validator instance.
     * @return \Cake\Validation\Validator
     */
    public function validationDefault(Validator $validator): Validator {
        $validator
            ->integer('container_id')
            ->requirePresence('container_id', 'create')
            ->notEmptyString('container_id');

        $validator
            ->scalar('name')
            ->maxLength('name', 200)
            ->requirePresence('name', 'create')
            ->notEmptyString('name');

        $validator
            ->scalar('description')
            ->maxLength('description', 255)
            ->allowEmptyString('description');

        $validator
            ->scalar('host_keywords')
            ->maxLength('host_keywords', 255)
            ->notEmptyString('host_keywords');

        $validator
            ->scalar('host_not_keywords')
            ->maxLength('host_not_keywords', 255)
            ->allowEmptyString('host_not_keywords');

        $validator
            ->scalar('service_keywords')
            ->maxLength('service_keywords', 255)
            ->allowEmptyString('service_keywords');

        $validator
            ->scalar('service_not_keywords')
            ->maxLength('service_not_keywords', 255)
            ->allowEmptyString('service_not_keywords');

        $validator
            ->scalar('servicename_regex')
            ->maxLength('servicename_regex', 2500)
            ->requirePresence('servicename_regex', 'create')
            ->allowEmptyString('servicename_regex')
            ->add('servicename_regex', 'custom', [
                'rule'    => [$this, 'isRegularExpression'],
                'message' => __('Your service regex is an invalid regular expression.')
            ]);

        $validator
            ->boolean('recursive')
            ->notEmptyString('recursive');

        return $validator;
    }

    /**
     * Returns a rules checker object that will be used for validating
     * application integrity.
     *
     * @param \Cake\ORM\RulesChecker $rules The rules object to be modified.
     * @return \Cake\ORM\RulesChecker
     */
    public function buildRules(RulesChecker $rules): RulesChecker {
        $rules->add($rules->existsIn('container_id', 'Containers'), ['errorField' => 'container_id']);

        return $rules;
    }

    /**
     * @param GenericFilter $GenericFilter
     * @param PaginateOMat|null $PaginateOMat
     * @param array $MY_RIGHTS
     * @return array
     */
    public function getCustomalertRulesIndex(GenericFilter $GenericFilter, PaginateOMat $PaginateOMat = null, $MY_RIGHTS = []) {
        $query = $this->find('all')
            ->disableHydration();


        $indexFilter = [];
        if (!empty($MY_RIGHTS)) {
            $indexFilter['CustomalertRules.container_id IN'] = $MY_RIGHTS;
        }
        $query->where($indexFilter);

        if (!empty($GenericFilter->genericFilters())) {
            $query->where($GenericFilter->genericFilters());
        }

        $query->order($GenericFilter->getOrderForPaginator('CustomalertRules.name', 'asc'));
        if ($PaginateOMat === null) {
            //Just execute query
            $result = $query->toArray();
        } else {
            if ($PaginateOMat->useScroll()) {
                $result = $this->scrollCake4($query, $PaginateOMat->getHandler());
            } else {
                $result = $this->paginateCake4($query, $PaginateOMat->getHandler());
            }
        }

        return $result;
    }

    /**
     * @param int $id
     * @return bool
     */
    public function existsById($id) {
        return $this->exists(['CustomalertRules.id' => $id]);
    }

    public function isRegularExpression($string) {
        return @preg_match('`' . $string . '`', '') !== false;
    }


    public function getServicesByCustomalertRule(ServicesTable $ServicesTable, CustomalertRule $customalertRule, array $containerIds) {
        $query = $ServicesTable->find()
            ->select([
                'keyField'   => 'Services.uuid',
                'valueField' => 'Services.id'
            ]);
        $query->join([
            [
                'table'      => 'hosts',
                'alias'      => 'Hosts',
                'type'       => 'INNER',
                'conditions' => [
                    'Services.host_id = Hosts.id'
                ],
            ],
            [
                'table'      => 'hosttemplates',
                'alias'      => 'Hosttemplates',
                'type'       => 'INNER',
                'conditions' => [
                    'Hosts.hosttemplate_id = Hosttemplates.id'
                ],
            ],
            [
                'table'      => 'servicetemplates',
                'alias'      => 'Servicetemplates',
                'type'       => 'INNER',
                'conditions' => [
                    'Services.servicetemplate_id = Servicetemplates.id'
                ],
            ],
        ]);
        $where = [
            'Services.disabled' => 0,
        ];

        if (!empty($containerIds)) {
            $where['Hosts.container_id IN'] = $containerIds;
        }

        $where[] = new ComparisonExpression(
            'IF((Hosts.tags IS NULL OR Hosts.tags=""), Hosttemplates.tags, Hosts.tags)',
            sprintf('.*(%s).*', implode('|', explode(',', $customalertRule->host_keywords))),
            'string',
            'RLIKE'
        );

        if (!empty($customalertRule->host_not_keywords)) {
            $where[] = new ComparisonExpression(
                'IF((Hosts.tags IS NULL OR Hosts.tags=""), Hosttemplates.tags, Hosts.tags)',
                sprintf('.*(%s).*', implode('|', explode(',', $customalertRule->host_not_keywords))),
                'string',
                'NOT RLIKE'
            );
        }

        if (!empty($customalertRule->service_keywords)) {
            $where[] = new ComparisonExpression(
                'IF((Services.tags IS NULL OR Hosts.tags=""), Servicetemplates.tags, Services.tags)',
                sprintf('.*(%s).*', implode('|', explode(',', $customalertRule->service_keywords))),
                'string',
                'RLIKE'
            );
        }

        if (!empty($customalertRule->service_not_keywords)) {
            $where[] = new ComparisonExpression(
                'IF((Services.tags IS NULL OR Hosts.tags=""), Servicetemplates.tags, Services.tags)',
                sprintf('.*(%s).*', implode('|', explode(',', $customalertRule->service_not_keywords))),
                'string',
                'NOT RLIKE'
            );
        }

        if (!empty($customalertRule->servicename_regex)) {
            $where[] = new ComparisonExpression(
                'IF((Services.name IS NULL OR Services.name=""), Servicetemplates.name, Services.name)',
                $customalertRule->servicename_regex,
                'string',
                'REGEXP'
            );
        }

        $query->where($where)
            ->disableHydration();

        return Hash::combine($query->all()->toList(), '{n}.keyField', '{n}.valueField');

    }

    /**
     * @param ServicesTable $ServicesTable
     * @param CustomalertRule $customalertRule
     * @param PaginateOMat $PaginateOMat
     * @param GenericFilter $GenericFilter
     * @param array $containerIds
     * @return array
     */
    public function getServicesForOverviewByCustomalertRule(ServicesTable $ServicesTable, CustomalertRule $customalertRule, PaginateOMat $PaginateOMat, GenericFilter $GenericFilter, array $containerIds) {
        $query = $ServicesTable->find();
        $query->select([
            'Hosts.name',
            'Hosts.id',
            'Services.id',
            'servicename' => $query->newExpr('IF((Services.name IS NULL OR Services.name=""), Servicetemplates.name, Services.name)'),
            'Services.description'
        ])
            ->join([
                [
                    'table'      => 'hosts',
                    'alias'      => 'Hosts',
                    'type'       => 'INNER',
                    'conditions' => [
                        'Services.host_id = Hosts.id'
                    ],
                ],
                [
                    'table'      => 'hosttemplates',
                    'alias'      => 'Hosttemplates',
                    'type'       => 'INNER',
                    'conditions' => [
                        'Hosts.hosttemplate_id = Hosttemplates.id'
                    ],
                ],
                [
                    'table'      => 'servicetemplates',
                    'alias'      => 'Servicetemplates',
                    'type'       => 'INNER',
                    'conditions' => [
                        'Services.servicetemplate_id = Servicetemplates.id'
                    ],
                ],
            ]);
        $where = [
            'Services.disabled'     => 0,
            'Hosts.container_id IN' => $containerIds,
            new ComparisonExpression(
                'IF((Hosts.tags IS NULL OR Hosts.tags=""), Hosttemplates.tags, Hosts.tags)',
                sprintf('.*(%s).*', implode('|', explode(',', $customalertRule->host_keywords))),
                'string',
                'RLIKE'
            )
        ];
        if (!empty($customalertRule->host_not_keywords)) {
            $where[] = new ComparisonExpression(
                'IF((Hosts.tags IS NULL OR Hosts.tags=""), Hosttemplates.tags, Hosts.tags)',
                sprintf('.*(%s).*', implode('|', explode(',', $customalertRule->host_not_keywords))),
                'string',
                'NOT RLIKE'
            );
        }

        if (!empty($customalertRule->service_keywords)) {
            $where[] = new ComparisonExpression(
                'IF((Services.tags IS NULL OR Hosts.tags=""), Servicetemplates.tags, Services.tags)',
                sprintf('.*(%s).*', implode('|', explode(',', $customalertRule->service_keywords))),
                'string',
                'RLIKE'
            );
        }

        if (!empty($customalertRule->service_not_keywords)) {
            $where[] = new ComparisonExpression(
                'IF((Services.tags IS NULL OR Hosts.tags=""), Servicetemplates.tags, Services.tags)',
                sprintf('.*(%s).*', implode('|', explode(',', $customalertRule->service_not_keywords))),
                'string',
                'NOT RLIKE'
            );
        }

        if (!empty($customalertRule->servicename_regex)) {
            $where[] = new ComparisonExpression(
                'IF((Services.name IS NULL OR Services.name=""), Servicetemplates.name, Services.name)',
                $customalertRule->servicename_regex,
                'string',
                'REGEXP'
            );
        }

        if (!empty($GenericFilter->genericFilters())) {
            $genericFilter = $GenericFilter->genericFilters();
            if (isset($genericFilter['servicename LIKE'])) {
                $where[] = new ComparisonExpression(
                    'IF((Services.name IS NULL OR Services.name=""), Servicetemplates.name, Services.name)',
                    $genericFilter['servicename LIKE'],
                    'string',
                    'LIKE'
                );
            }
            if (isset($genericFilter['Hosts.name LIKE'])) {
                $where[] = new ComparisonExpression(
                    'Hosts.name',
                    $genericFilter['Hosts.name LIKE'],
                    'string',
                    'LIKE'
                );
            }
        }

        $query->where($where);

        $query->order(
            array_merge(
                $GenericFilter->getOrderForPaginator('Hosts.name', 'asc'),
                ['Services.id' => 'asc']
            )
        );

        $query->disableHydration();


        if ($PaginateOMat === null) {
            //Just execute query
            $result = $this->emptyArrayIfNull($query->toArray());
        } else {
            if ($PaginateOMat->useScroll()) {
                $result = $this->scrollCake4($query, $PaginateOMat->getHandler());
            } else {
                $result = $this->paginateCake4($query, $PaginateOMat->getHandler());
            }
        }
        return $result;
    }
}
