<?php
// Copyright (C) <2015-present>  <it-novum GmbH>
//
// This file is licensed under the terms of the openITCOCKPIT Enterprise Edition license agreement.
// The license agreement and license key were sent with the order confirmation.

declare(strict_types=1);

namespace CustomalertModule\Model\Table;

use App\Lib\Traits\PaginationAndScrollIndexTrait;
use Cake\ORM\Behavior\TimestampBehavior;
use Cake\ORM\Query;
use Cake\ORM\RulesChecker;
use Cake\ORM\Table;
use Cake\ORM\TableRegistry;
use Cake\Validation\Validator;
use CustomalertModule\Lib\CustomalertConditions;
use CustomalertModule\Model\Entity\Customalert;
use itnovum\openITCOCKPIT\Database\PaginateOMat;

/**
 * Customalerts Model
 *
 * @property ServicesTable&\Cake\ORM\Association\BelongsTo $Services
 * @property CustomalertCommentsTable&\Cake\ORM\Association\HasMany $CustomalertComments
 * @property CustomalertStatehistoryTable&\Cake\ORM\Association\HasMany $CustomalertStatehistory
 *
 * @method Customalert newEmptyEntity()
 * @method Customalert newEntity(array $data, array $options = [])
 * @method Customalert[] newEntities(array $data, array $options = [])
 * @method Customalert get($primaryKey, $options = [])
 * @method Customalert findOrCreate($search, ?callable $callback = null, $options = [])
 * @method Customalert patchEntity(\Cake\Datasource\EntityInterface $entity, array $data, array $options = [])
 * @method Customalert[] patchEntities(iterable $entities, array $data, array $options = [])
 * @method Customalert|false save(\Cake\Datasource\EntityInterface $entity, $options = [])
 * @method Customalert saveOrFail(\Cake\Datasource\EntityInterface $entity, $options = [])
 * @method Customalert[]|\Cake\Datasource\ResultSetInterface|false saveMany(iterable $entities, $options = [])
 * @method Customalert[]|\Cake\Datasource\ResultSetInterface saveManyOrFail(iterable $entities, $options = [])
 * @method Customalert[]|\Cake\Datasource\ResultSetInterface|false deleteMany(iterable $entities, $options = [])
 * @method Customalert[]|\Cake\Datasource\ResultSetInterface deleteManyOrFail(iterable $entities, $options = [])
 *
 * @mixin TimestampBehavior
 */
class CustomalertsTable extends Table {
    use PaginationAndScrollIndexTrait;

    /**
     * Initialize method
     *
     * @param array $config The configuration for the Table.
     * @return void
     */
    public function initialize(array $config): void {
        parent::initialize($config);

        $this->setTable('customalerts');
        $this->setDisplayField('id');
        $this->setPrimaryKey('id');

        $this->addBehavior('Timestamp');

        $this->belongsTo('Services', [
            'foreignKey' => 'service_id',
            'joinType'   => 'INNER',
            'className'  => 'Services',
        ]);
        $this->hasMany('CustomalertComments', [
            'foreignKey' => 'customalert_id',
            'className'  => 'CustomalertModule.CustomalertComments',
        ]);
        $this->hasMany('CustomalertStatehistory', [
            'foreignKey' => 'customalert_id',
            'className'  => 'CustomalertModule.CustomalertStatehistory',
        ]);
    }

    /**
     * Default validation rules.
     *
     * @param Validator $validator Validator instance.
     * @return Validator
     */
    public function validationDefault(Validator $validator): Validator {
        $validator
            ->scalar('service_id')
            ->maxLength('service_id', 37)
            ->requirePresence('service_id', 'create')
            ->notEmptyString('service_id');

        $validator
            ->scalar('service_uuid')
            ->maxLength('service_uuid', 37)
            ->requirePresence('service_uuid', 'create')
            ->notEmptyString('service_uuid');

        // Also change $maxLength in CustomalertModule/src/Command/CustomAlertWorkerCommand.php if you change the length
        $validator
            ->scalar('message')
            ->maxLength('message', 1024)
            ->requirePresence('message', 'create')
            ->notEmptyString('message');

        $validator
            ->nonNegativeInteger('state')
            ->allowEmptyString('state');

        return $validator;
    }

    /**
     * Returns a rules checker object that will be used for validating
     * application integrity.
     *
     * @param RulesChecker $rules The rules object to be modified.
     * @return RulesChecker
     */
    public function buildRules(RulesChecker $rules): RulesChecker {
        $rules->add($rules->existsIn('service_id', 'Services'), ['errorField' => 'service_id']);

        return $rules;
    }

    /**
     * @param CustomalertConditions $CustomalertConditions
     * @param PaginateOMat|null $PaginateOMat
     * @param $MY_RIGHTS
     * @return mixed
     */
    public function getCustomalertsIndex(CustomalertConditions $CustomalertConditions, PaginateOMat $PaginateOMat = null, $MY_RIGHTS = []) {
        $query = $this->find();
        $query
            ->select([
                'Customalerts.id',
                'Customalerts.message',
                'Customalerts.state',
                'Customalerts.latest_comment_id',
                'Customalerts.service_id',
                'Customalerts.created',
                'Customalerts.modified',
                'servicename' => $query->newExpr('IF((Services.name IS NULL OR Services.name=""), Servicetemplates.name, Services.name)'),
                'CustomalertComments.comment',
                'CustomalertComments.entry_time',
            ])
            ->contain([
                'Services' => function (Query $query) use ($MY_RIGHTS) {
                    $query->disableAutoFields()
                        ->select([
                            'Services.id',
                            'Services.host_id',
                            'Services.disabled'
                        ])
                        ->where(['Services.disabled' => 0])
                        ->contain([
                            'Hosts'            => function (Query $query) use ($MY_RIGHTS) {
                                $query->disableAutoFields();

                                $query->select([
                                    'Hosts.id',
                                    'Hosts.name',
                                    'container' => $query->newExpr('SELECT Containers.name')
                                ]);
                                if (!empty($MY_RIGHTS)) {
                                    $query->select([
                                        'primary_container' => $query->newExpr(
                                            'SELECT IF(Containers.name = "root", "root", CONCAT(
                                            (
                                                SELECT GROUP_CONCAT(containers_temp.name ORDER BY containers_temp.lft ASC SEPARATOR "/")
                                                FROM containers AS containers_orig
                                                LEFT JOIN containers AS containers_temp ON containers_temp.lft < containers_orig.lft AND containers_temp.rght > containers_orig.rght
                                                WHERE containers_orig.id = Containers.id
                                                AND containers_temp.id IN (' . implode(',', $MY_RIGHTS) . ')
                                                ORDER BY containers_orig.lft ASC
                                            ), "/",
                                            Containers.name)
                                            )
                                        ')
                                    ]);

                                } else {
                                    $query->select([
                                        'primary_container' => $query->newExpr(
                                            'SELECT IF(Containers.name = "root", "root", CONCAT(
                                            (
                                                SELECT GROUP_CONCAT(containers_temp.name ORDER BY containers_temp.lft ASC SEPARATOR "/")
                                                FROM containers AS containers_orig
                                                LEFT JOIN containers AS containers_temp ON containers_temp.lft < containers_orig.lft AND containers_temp.rght > containers_orig.rght
                                                WHERE containers_orig.id = Containers.id
                                                ORDER BY containers_orig.lft ASC
                                            ), "/",
                                            Containers.name)
                                            )
                                        ')
                                    ]);
                                }
                                $query->innerJoinWith('HostsToContainersSharing', function (Query $q) use ($MY_RIGHTS) {
                                    if (!empty($MY_RIGHTS)) {
                                        return $q->where(['HostsToContainersSharing.id IN' => $MY_RIGHTS]);
                                    }
                                    return $q;
                                })->innerJoinWith('Containers');
                                return $query;
                            },
                            'Servicetemplates' => function (Query $query) {
                                $query->disableAutoFields()
                                    ->select([
                                        'Servicetemplates.id'
                                    ]);
                                return $query;
                            }
                        ]);
                    return $query;
                }
            ]);

        $query->join([
            [
                'table'      => 'customalert_comments',
                'alias'      => 'CustomalertComments',
                'type'       => 'LEFT',
                'conditions' => [
                    'CustomalertComments.id = Customalerts.latest_comment_id',
                ],
            ]
        ]);

        $query->where([
            'Customalerts.created >' => date('Y-m-d H:i:s', $CustomalertConditions->getFrom()),
            'Customalerts.created <' => date('Y-m-d H:i:s', $CustomalertConditions->getTo())
        ]);

        if ($CustomalertConditions->hasConditions()) {
            $query->andWhere($CustomalertConditions->getConditions());
        }

        $query->disableHydration();
        $query->group(['Customalerts.id']);
        $query->order($CustomalertConditions->getOrder());

        if ($PaginateOMat === null) {
            //Just execute query
            $result = $this->emptyArrayIfNull($query->toArray());
        } else {
            if ($PaginateOMat->useScroll()) {
                $result = $this->scrollCake4($query, $PaginateOMat->getHandler());
            } else {
                $result = $this->paginateCake4($query, $PaginateOMat->getHandler());
            }
        }

        return $result;
    }

    /**
     * @param int $serviceId
     * @param string $message
     * @return int
     */
    public function getLast24hCounter(int $serviceId, string $message) {
        $count = $this->find()
            ->where([
                'service_id' => $serviceId,
                'message'    => $message,
                'created >= NOW() - INTERVAL 1 DAY'
            ])
            ->count();

        return $count;
    }

    /**
     * @param $id
     * @return bool
     */
    public function existsById($id) {
        return $this->exists(['Customalerts.id' => $id]);
    }

    /**
     * @param int $serviceId
     * @param int[] $states
     * @return Customalert[]
     */
    public function getCustomalertsForWorker(int $serviceId, array $states) {
        $query = $this->find()
            ->where([
                'service_id' => $serviceId,
                'state IN'   => $states
            ])
            ->all();

        return $query->toArray();
    }

    /**
     * @param Customalert $customalert
     * @param int $state
     * @param int $userId
     * @param int|null $commentId
     * @return bool
     */
    public function changeState(Customalert $customalert, int $state, ?int $commentId = null): bool {
        $customalert->set('state', $state);
        if (!empty($commentId)) {
            $customalert->set('latest_comment_id', $commentId);
        }
        if ($this->save($customalert)) {
            /** @var CustomalertStatehistoryTable $CustomalertStatehistoryTable */
            $CustomalertStatehistoryTable = TableRegistry::getTableLocator()->get('CustomalertModule.CustomalertStatehistory');
            $history = $CustomalertStatehistoryTable->newEntity([
                'customalert_id'         => $customalert->id,
                'state'                  => $customalert->state,
                'state_time'             => $customalert->modified,
                'customalert_comment_id' => $commentId
            ]);

            if ($CustomalertStatehistoryTable->save($history) !== false) {
                return true;
            }
        }

        return false;
    }

    /**
     * @param Customalert $customalert
     * @return array|bool
     */
    public function createCustomalert(Customalert $customalert) {
        $this->save($customalert);
        if ($customalert->hasErrors()) {
            return $customalert->getErrors();
        }

        /** @var CustomalertStatehistoryTable $CustomalertStatehistoryTable */
        $CustomalertStatehistoryTable = TableRegistry::getTableLocator()->get('CustomalertModule.CustomalertStatehistory');
        $history = $CustomalertStatehistoryTable->newEntity([
            'customalert_id'         => $customalert->id,
            'state'                  => $customalert->state,
            'state_time'             => $customalert->created,
            'customalert_comment_id' => null
        ]);

        if ($CustomalertStatehistoryTable->save($history) !== false) {
            return true;
        }

        return false;
    }

    /**
     * @param Customalert $customalert
     * @return bool
     */
    public function closeCustomalert(Customalert $customalert) {
        $customalert->set('state', Customalert::STATE_DONE);
        $this->save($customalert);
        if (!$customalert->hasErrors()) {
            /** @var CustomalertStatehistoryTable $CustomalertStatehistoryTable */
            $CustomalertStatehistoryTable = TableRegistry::getTableLocator()->get('CustomalertModule.CustomalertStatehistory');
            $history = $CustomalertStatehistoryTable->newEntity([
                'customalert_id'         => $customalert->id,
                'state'                  => Customalert::STATE_DONE,
                'state_time'             => date('Y-m-d H:i:s'),
                'customalert_comment_id' => null
            ]);

            if ($CustomalertStatehistoryTable->save($history) !== false) {
                return true;
            }
        }

        return false;
    }

    /**
     * @param int $customalertId
     * @return array
     */
    public function getCustomalertHistoryById($customalertId) {
        $result = $this->find()
            ->contain([
                    'Services' => function (Query $q) {
                        return $q->enableAutoFields(false)
                            ->select([
                                'Hosts.id',
                                'Hosts.name',
                                'Services.id',
                                'servicename' => $q->newExpr('IF((Services.name IS NULL OR Services.name=""), Servicetemplates.name, Services.name)'),
                            ])
                            ->contain([
                                'Hosts'            => function (Query $q) {
                                    return $q->enableAutoFields(false)
                                        ->select([
                                            'Hosts.id',
                                            'Hosts.name',
                                            'Hosts.container_id'
                                        ]);
                                },
                                'Servicetemplates' => function (Query $q) {
                                    return $q->enableAutoFields(false)
                                        ->select([
                                            'Servicetemplates.id'
                                        ]);
                                }
                            ]);
                    }
                ]
            )
            ->contain([
                'CustomalertStatehistory' => function (Query $q) {
                    return $q->enableAutoFields(false)
                        ->contain([
                            'CustomalertComments' => [
                                'Users' => function (Query $q) {
                                    return $q->enableAutoFields(false)
                                        ->select([
                                            'Users.id',
                                            'Users.firstname',
                                            'Users.lastname'
                                        ]);
                                }
                            ]
                        ])->order([
                            'CustomalertStatehistory.state_time' => 'desc'
                        ]);
                }
            ])
            ->where([
                'Customalerts.id' => $customalertId
            ])
            ->disableAutoFields()
            ->disableHydration()
            ->first();

        if (empty($result)) {
            return [];
        }
        return $result;
    }

    /**
     * @param int $serviceId
     * @param null|PaginateOMat $PaginateOMat
     * @return array|\Cake\Datasource\EntityInterface
     */
    public function getCustomalertHistoryByServiceId($serviceId, $PaginateOMat = null) {
        $query = $this->find()
            ->contain([
                'CustomalertStatehistory' => function (Query $q) {
                    return $q->enableAutoFields(false)
                        ->contain([
                            'CustomalertComments' => [
                                'Users' => function (Query $q) {
                                    return $q->enableAutoFields(false)
                                        ->select([
                                            'Users.id',
                                            'Users.firstname',
                                            'Users.lastname'
                                        ]);
                                }
                            ]
                        ])->where([
                            'CustomalertStatehistory.state > 0'
                        ])
                        ->order([
                            'CustomalertStatehistory.state_time' => 'asc',
                            'CustomalertStatehistory.id'         => 'asc'
                        ]);
                }
            ])
            ->where([
                'Customalerts.service_id' => $serviceId
            ])
            ->order([
                'Customalerts.modified' => 'desc',
                'Customalerts.id'       => 'desc'
            ])
            ->disableAutoFields()
            ->disableHydration();

        if ($PaginateOMat === null) {
            //Just execute query
            $result = $this->emptyArrayIfNull($query->toArray());
        } else {
            if ($PaginateOMat->useScroll()) {
                $result = $this->scrollCake4($query, $PaginateOMat->getHandler());
            } else {
                $result = $this->paginateCake4($query, $PaginateOMat->getHandler());
            }
        }

        return $result;
    }

    /**
     * @param $serviceId
     * @return bool
     */
    public function existsCustomalertsByServiceId($serviceId) {
        return !empty($this->find()
            ->where([
                'Customalerts.service_id' => $serviceId
            ])
            ->first()
        );
    }

    /**
     * @param CustomalertConditions $CustomalertConditions
     * @param $MY_RIGHTS
     * @return array
     */
    public function getCustomalertsGlobalOverviewNewAndInProgress(CustomalertConditions $CustomalertConditions, $MY_RIGHTS = []): array {
        $query = $this->find();
        $query->select([
            'Customalerts.service_id',
            'Customalerts.state',
            'count' => $query->newExpr('COUNT(DISTINCT Customalerts.id)'),
        ])
            ->contain([
                'Services' => function (Query $query) use ($MY_RIGHTS) {
                    $query->disableAutoFields()
                        ->where(['Services.disabled' => 0])
                        ->contain([
                            'Hosts'            => function (Query $query) use ($MY_RIGHTS) {
                                $query->disableAutoFields();
                                $query->innerJoinWith('HostsToContainersSharing', function (Query $q) use ($MY_RIGHTS) {
                                    if (!empty($MY_RIGHTS)) {
                                        return $q->where(['HostsToContainersSharing.id IN' => $MY_RIGHTS]);
                                    }
                                    return $q;
                                })->innerJoinWith('Containers');
                                return $query;
                            },
                            'Servicetemplates' => function (Query $query) {
                                $query->disableAutoFields();
                                return $query;
                            }
                        ]);
                    return $query;
                }
            ]);

        $query->join([
            [
                'table'      => 'customalert_comments',
                'alias'      => 'CustomalertComments',
                'type'       => 'LEFT',
                'conditions' => [
                    'CustomalertComments.id = Customalerts.latest_comment_id',
                ],
            ]
        ]);

        $query->where([
            'Customalerts.state IN'  => [Customalert::STATE_NEW, Customalert::STATE_PROGRESS],
            'Customalerts.created >' => date('Y-m-d H:i:s', $CustomalertConditions->getFrom()),
            'Customalerts.created <' => date('Y-m-d H:i:s', $CustomalertConditions->getTo())
        ]);

        if ($CustomalertConditions->hasConditions()) {
            $query->andWhere($CustomalertConditions->getConditions());
        }
        $query->count();

        $query->disableHydration();
        $query->group(['Customalerts.state']);


        return $this->emptyArrayIfNull($query->toArray());
    }
}
